#! /usr/bin/env python
# -*- coding: utf-8 -*-

import pandas as pd

from hypothesis import given, settings

from saqc.constants import UNFLAGGED, BAD
from saqc.core.register import FunctionWrapper

from tests.fuzzy.lib import dataFieldFlags, MAX_EXAMPLES


@settings(max_examples=MAX_EXAMPLES, deadline=None)
@given(data_field_flags=dataFieldFlags())
def test_maskingMasksData(data_field_flags):
    """
    test if flagged values are replaced by np.nan
    """
    data_in, field, flags = data_field_flags
    data_masked, mask = FunctionWrapper._maskData(
        data_in, flags, columns=[field], thresh=UNFLAGGED
    )  # thresh UNFLAGGED | np.inf
    assert data_masked[field].iloc[mask[field].index].isna().all()
    assert (flags[field].iloc[mask[field].index] > UNFLAGGED).all()


@settings(max_examples=MAX_EXAMPLES, deadline=None)
@given(data_field_flags=dataFieldFlags())
def test_dataMutationPreventsUnmasking(data_field_flags):
    """test if (un)masking works as expected on data-changes.

    if `data` is mutated after `_maskData`, `_unmaskData` should be a no-op
    """
    filler = -9999

    data_in, field, flags = data_field_flags

    data_masked, mask = FunctionWrapper._maskData(
        data_in, flags, columns=[field], thresh=UNFLAGGED
    )
    data_masked[field] = filler
    data_out = FunctionWrapper._unmaskData(data_masked, mask)
    assert (data_out[field] == filler).all(axis=None)


@settings(max_examples=MAX_EXAMPLES, deadline=None)
@given(data_field_flags=dataFieldFlags())
def test_flagsMutationPreventsUnmasking(data_field_flags):
    """test if (un)masking works as expected on flags-changes.

    if `flags` is mutated after `_maskData`, `_unmaskData` should be a no-op
    """
    data_in, field, flags = data_field_flags

    data_masked, mask = FunctionWrapper._maskData(
        data_in, flags, columns=[field], thresh=UNFLAGGED
    )
    flags[:, field] = UNFLAGGED
    data_out = FunctionWrapper._unmaskData(data_masked, mask)
    assert (data_out.loc[flags[field] == BAD, field].isna()).all(axis=None)


@settings(max_examples=MAX_EXAMPLES, deadline=None)
@given(data_field_flags=dataFieldFlags())
def test_reshapingPreventsUnmasking(data_field_flags):
    """test if (un)masking works as expected on index-changes.

    If the index of data (and flags) change in the func, the unmasking,
    should not reapply original data, instead take the new data (and flags) as is.
    """

    filler = -1111

    data_in, field, flags = data_field_flags

    data_masked, mask = FunctionWrapper._maskData(
        data_in, flags, columns=[field], thresh=UNFLAGGED
    )
    # mutate indexes of `data` and `flags`
    index = data_masked[field].index.to_series()
    index.iloc[-len(data_masked[field]) // 2 :] += pd.Timedelta("7.5Min")
    data_masked[field] = pd.Series(data=filler, index=index)

    fflags = flags[field]
    flags.drop(field)
    flags[field] = pd.Series(data=fflags.values, index=index)

    data_out = FunctionWrapper._unmaskData(data_masked, mask)
    assert (data_out[field] == filler).all(axis=None)


@settings(max_examples=MAX_EXAMPLES, deadline=None)
@given(data_field_flags=dataFieldFlags())
def test_unmaskingInvertsMasking(data_field_flags):
    """
    unmasking data should invert the masking
    """
    data_in, field, flags = data_field_flags

    data_masked, mask = FunctionWrapper._maskData(
        data_in, flags, columns=[field], thresh=UNFLAGGED
    )
    data_out = FunctionWrapper._unmaskData(data_masked, mask)
    assert pd.DataFrame.equals(
        data_out.to_df().astype(float), data_in.to_df().astype(float)
    )


# @settings(max_examples=MAX_EXAMPLES, deadline=None)
# @given(data_field_flags=dataFieldFlags(), func_kwargs=flagFuncsKwargs())
# def test_maskingPreservesData(data_field_flags, func_kwargs):
#     """
#     no mutations on pre-flagged data

#     calling a function on pre-flagged data should yield the same
#     behavior as calling this function on data where the flagged values
#     are removed
#     """

#     data_in, field, flags = data_field_flags

#     data_masked, mask = _maskData(data_in, flags, columns=[field], dfilter=flags.BAD)
#     func, kwargs = func_kwargs
#     data_masked, _ = func(data_masked, field, flags, **kwargs)
#     data_out = _unmaskData(data_in, mask, data_masked, flags, dfilter=flags.BAD)

#     flags_in = flags.isFlagged(flag=flags.BAD)
#     assert data_in.aloc[flags_in].equals(data_out.aloc[flags_in])


# @settings(max_examples=MAX_EXAMPLES, deadline=None)
# @given(data_field_flags=dataFieldFlags(), func_kwargs=flagFuncsKwargs())
# def test_maskingEqualsRemoval(data_field_flags, func_kwargs):
#     """
#     calling a function on pre-flagged data should yield the same
#     results as calling this function on data where the flagged values
#     are removed
#     """
#     func, kwargs = func_kwargs

#     data, field, flags = data_field_flags
#     flagged_in = flags.isFlagged(flag=flags.BAD, comparator=">=")

#     # mask and call
#     data_left, _ = _maskData(data, flags, columns=[field], dfilter=flags.BAD)
#     data_left, _ = func(data_left, field, flags, **kwargs)

#     # remove and call
#     data_right = data.aloc[~flagged_in]
#     flags_right = flags.initFlags(flags.getFlags().aloc[~flagged_in])
#     data_right, _ = func(data_right, field, flags_right, **kwargs)

#     # NOTE: we need to handle the implicit type conversion in `_maskData`
#     data_left_compare = data_left.aloc[~flagged_in]
#     data_left_compare[field] = data_left_compare[field].astype(data[field].dtype)

#     assert data_right.equals(data_left_compare)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import functools
from typing import Callable, Any

# from typing_extensions import ParamSpec

# P = ParamSpec("P")

# def wrapper(saqc: "SaQC", func: Callable[P, int]) -> Callable[P, "SaQC"]:
#     def inner(*args: P.args, **kwargs: P.kwargs) -> "SaQC":
#         return saqc
#     return inner

# def func(x: int, y: int) -> int:
#     """NotMuch"""
#     return x + y

# FUNC_MAP = {"func": func}

# class FuncDescriptor:
#     def __init__(self, callable):
#         self.f = callable
#         self.annotation = Callable[[float, float], float]

#     def __get__(self, obj, type=None):
#         return self.f

#     # @property
#     # def __annotations__(self):
#     #     return {**self.f.__annotations__, "return": "SaQC"}

#     @property
#     def __func__(self):
#         return self.f


# class SaQC:
#     def __init__(self):
#         self._planned = []

#     # def _wrap(self, func_name):
#     #     def inner(*args, **kwargs):
#     #         self._planned.append((func_name, args, kwargs))
#     #         return self
#     #     return inner

#     def _wrap(self, func_name, kwargs):
#         self._planned.append((func_name, kwargs))
#         return self

#     # def __getattr__(self, key):
#     #     if key not in FUNC_MAP:
#     #         raise AttributeError(f"no such attribute: '{key}'")
#     #     return self._wrap(key)

#     # @functools.wraps(func)
#     def funcMethod(self, x: int, y: int) -> "SaQC":
#         self._wrap("func", locals())
#         return self

#     def wrapper(self, *args, **kwargs):
#         self.wrapper.__annotations__.update({**func.__annotations__, "return": "SaQC"})
#         self._planned.append((func, args, kwargs))
#         return self


#     # fff = property(fget=lambda *args, **kwargs: func(*args, **kwargs))
#     ggg  = FuncDescriptor(func)
#     gggg = func

if __name__ == "__main__":
    from saqc.funcs.breaks import flagMissing
    import pdb

    pdb.set_trace()
    # saqc = SaQC()
    # xx = saqc.funcMethod
    # saqc.ggg(1)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Union, Callable, Tuple
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.constants import *
from saqc.core import register, Flags
from saqc.lib.tools import getFreqDelta, filterKwargs


@register(mask=["field"], demask=[], squeeze=[])
def roll(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    func: Callable[[pd.Series], np.ndarray] = np.mean,
    min_periods: int = 0,
    center: bool = True,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Calculate a rolling-window function on the data.

    Note, that the data gets assigned the worst flag present in the original data.

    Parameters
    ----------
    data : dios.DictOfSeries
        The data.
    field : str
        The column to calculate on.
    flags : saqc.Flags
        Container to store quality flags to data.
    window : {int, str}
        The size of the window you want to roll with. If an integer is passed, the size
        refers to the number of periods for every fitting window. If an offset string
        is passed, the size refers to the total temporal extension. For regularly
        sampled timeseries, the period number will be casted down to an odd number if
        ``center=True``.
    func : Callable, default np.mean
        Function to roll with.
    min_periods : int, default 0
        The minimum number of periods to get a valid value
    center : bool, default True
        If True, center the rolling window.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    reserved = ["return_residues", "set_flags"]
    kwargs = filterKwargs(kwargs, reserved)
    return _roll(
        data=data,
        field=field,
        flags=flags,
        window=window,
        func=func,
        min_periods=min_periods,
        center=center,
        **kwargs,
        # ctrl args
        set_flags=True,
        return_residues=False,
    )


def _roll(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    window: Union[str, int],
    func: Callable[[pd.Series], np.ndarray] = np.mean,
    set_flags: bool = True,
    min_periods: int = 0,
    center: bool = True,
    return_residues=False,
    **kwargs
):
    to_fit = data[field].copy()
    if to_fit.empty:
        return data, flags

    regular = getFreqDelta(to_fit.index)
    # starting with the annoying case: finding the rolling interval
    # centers of not-harmonized input time series:
    if center and not regular:
        if isinstance(window, int):
            raise NotImplementedError(
                "Integer based window size is not supported for not-harmonized"
                'sample series when rolling with "center=True".'
            )
        # get interval centers
        centers = np.floor(
            (
                to_fit.rolling(
                    pd.Timedelta(window) / 2, closed="both", min_periods=min_periods
                ).count()
            )
        )
        centers = centers.drop(centers[centers.isna()].index)
        centers = centers.astype(int)
        roller = to_fit.rolling(
            pd.Timedelta(window), closed="both", min_periods=min_periods
        )
        try:
            means = getattr(roller, func.__name__)()
        except AttributeError:
            means = to_fit.rolling(
                pd.Timedelta(window), closed="both", min_periods=min_periods
            ).apply(func)

        def center_func(x, y=centers):
            pos = x.index[int(len(x) - y[x.index[-1]])]
            return y.index.get_loc(pos)

        centers_iloc = (
            centers.rolling(window, closed="both")
            .apply(center_func, raw=False)
            .astype(int)
        )
        temp = means.copy()
        for k in centers_iloc.iteritems():
            means.iloc[k[1]] = temp[k[0]]
        # last values are false, due to structural reasons:
        means[means.index[centers_iloc[-1]] : means.index[-1]] = np.nan

    # everything is more easy if data[field] is harmonized:
    else:
        if isinstance(window, str):
            window = pd.Timedelta(window) // regular
        if (window % 2 == 0) & center:
            window = int(window - 1)

        roller = to_fit.rolling(window=window, center=center, closed="both")
        try:
            means = getattr(roller, func.__name__)()
        except AttributeError:
            means = to_fit.rolling(window=window, center=center, closed="both").apply(
                func
            )

    if return_residues:
        means = to_fit - means

    data[field] = means
    if set_flags:
        worst = flags[field].rolling(window, center=True, min_periods=min_periods).max()
        flags[field] = worst

    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional
from typing_extensions import Literal
from saqc.lib.tools import toSequence
import pandas as pd
import numpy as np
import matplotlib as mpl
import itertools
import matplotlib.pyplot as plt
import seaborn
from saqc.constants import *
from saqc.core import Flags
from saqc.lib.types import DiosLikeT


STATSDICT = {
    "values total": lambda x, y, z: len(x),
    "invalid total (=NaN)": lambda x, y, z: x.isna().sum(),
    "invalid percentage": lambda x, y, z: round((x.isna().sum()) / len(x), 2),
    "flagged total": lambda x, y, z: (y >= z).sum(),
    "flagged percentage": lambda x, y, z: round(((y >= z).sum()) / len(x), 2),
}

PLOT_KWARGS = {"alpha": 0.8, "linewidth": 1}
AX_KWARGS = {}
FIG_KWARGS = {"figsize": (16, 9)}
SCATTER_KWARGS = {
    "marker": ["s", "D", "^", "o"],
    "color": seaborn.color_palette("bright"),
    "alpha": 0.7,
    "zorder": 10,
    "edgecolors": "black",
    "s": 70,
}


def makeFig(
    data: DiosLikeT,
    field: str,
    flags: Flags,
    level: float,
    max_gap: Optional[str] = None,
    stats: bool = False,
    history: Optional[Literal["valid", "complete"]] = "valid",
    xscope: Optional[slice] = None,
    phaseplot: Optional[str] = None,
    stats_dict: Optional[dict] = None,
):
    """
    Returns a figure object, containing data graph with flag marks for field.

    Parameters
    ----------
    data : {pd.DataFrame, dios.DictOfSeries}
        data

    flags : {pd.DataFrame, dios.DictOfSeries, saqc.flagger}
        Flags or flagger object

    field : str
        Name of the variable-to-plot

    level : str, float, default None
        Flaglevel above wich flagged values should be displayed.

    max_gap : str, default None
        If None, all the points in the data will be connected, resulting in long linear
        lines, where continous chunks of data is missing. Nans in the data get dropped
        before plotting. If an Offset string is passed, only points that have a distance
        below `max_gap` get connected via the plotting line.

    stats : bool, default False
        Whether to include statistics table in plot.

    history : {"valid", "complete", None}, default "valid"
        Discriminate the plotted flags with respect to the tests they originate from.
        * "valid" - Only plot those flags, that do not get altered or "unflagged" by subsequent tests. Only list tests
          in the legend, that actually contributed flags to the overall resault.
        * "complete" - plot all the flags set and list all the tests ran on a variable. Suitable for debugging/tracking.
        * "clear" - clear plot from all the flagged values
        * None - just plot the resulting flags for one variable, without any historical meta information.

    xscope : slice or Offset, default None
        Parameter, that determines a chunk of the data to be plotted /
        processed. `s` can be anything, that is a valid argument to the ``pandas.Series.__getitem__`` method.

    phaseplot :

    stats_dict: dict, default None
        (Only relevant if `stats`=True).
        Dictionary of additional statisticts to write to the statistics table
        accompanying the data plot. An entry to the stats_dict has to be of the form:

        * {"stat_name": lambda x, y, z: func(x, y, z)}

        The lambda args ``x``,``y``,``z`` will be fed by:

        * ``x``: the data (``data[field]``).
        * ``y``: the flags (``flags[field]``).
        * ``z``: The passed flags level (``kwargs[flag]``)

        See examples section for examples


    Returns
    -------
    fig : matplotlib.pyplot.figure
        figure object.

    Examples
    --------
    Summary statistic function examples:

    >>> func = lambda x, y, z: len(x)

    Total number of nan-values:

    >>> func = lambda x, y, z: x.isna().sum()

    Percentage of values, flagged greater than passed flag (always round float results to avoid table cell overflow):

    >>> func = lambda x, y, z: round((x.isna().sum()) / len(x), 2)
    """

    if stats_dict is None:
        stats_dict = {}

    # data retrieval
    d = data[field]
    # data slicing:
    xscope = xscope or slice(xscope)
    d = d[xscope]
    flags_vals = flags[field][xscope]
    flags_hist = flags.history[field].hist.loc[xscope]
    flags_meta = flags.history[field].meta
    if stats:
        stats_dict.update(STATSDICT)
        stats_dict = _evalStatsDict(stats_dict, d, flags_vals, level)

    na_mask = d.isna()
    d = d[~na_mask]
    if phaseplot:
        flags_vals = flags_vals.copy()
        flags_hist = flags_hist.copy()
        phase_index = data[phaseplot][xscope].values
        phase_index_d = phase_index[~na_mask]
        na_mask.index = phase_index
        d.index = phase_index_d
        flags_vals.index = phase_index
        flags_hist.index = phase_index
        plot_kwargs = {**PLOT_KWARGS, **{"marker": "o", "linewidth": 0}}
        ax_kwargs = {**{"xlabel": phaseplot, "ylabel": d.name}, **AX_KWARGS}
    else:
        plot_kwargs = PLOT_KWARGS
        ax_kwargs = AX_KWARGS

    # insert nans between values mutually spaced > max_gap
    if max_gap and not d.empty:
        d = _insertBlockingNaNs(d, max_gap)

    # figure composition
    fig = mpl.pyplot.figure(constrained_layout=True, **FIG_KWARGS)
    grid = fig.add_gridspec()
    if stats:
        plot_gs, tab_gs = grid[0].subgridspec(ncols=2, nrows=1, width_ratios=[5, 1])
        ax = fig.add_subplot(tab_gs)
        _plotStatsTable(ax, stats_dict)
        ax = fig.add_subplot(plot_gs)
    else:
        ax = fig.add_subplot(grid[0])

    _plotVarWithFlags(
        ax,
        d,
        flags_vals,
        flags_hist,
        flags_meta,
        history,
        level,
        na_mask,
        plot_kwargs,
        ax_kwargs,
        SCATTER_KWARGS,
    )
    return fig


def _evalStatsDict(in_dict, datser, flagser, level):
    out_dict = {}
    for key in in_dict:
        out_dict[key] = str(in_dict[key](datser, flagser, level))
    return out_dict


def _plotStatsTable(ax, stats_dict):
    ax.axis("tight")
    ax.axis("off")
    tab_obj = ax.table(
        cellText=[[a, b] for a, b in stats_dict.items()],
        in_layout=True,
        loc="center",
        # make the table a bit smaller than the plot
        bbox=[0.0, 0.1, 0.95, 0.8],
    )
    tab_obj.auto_set_column_width(False)
    tab_obj.auto_set_font_size(False)
    tab_obj.set_fontsize(10)


def _plotVarWithFlags(
    ax,
    datser,
    flags_vals,
    flags_hist,
    flags_meta,
    history,
    level,
    na_mask,
    plot_kwargs,
    ax_kwargs,
    scatter_kwargs,
):
    scatter_kwargs = scatter_kwargs.copy()
    ax.set_title(datser.name)
    ax.plot(datser, color="black", **plot_kwargs)
    ax.set(**ax_kwargs)
    shape_cycle = scatter_kwargs.get("marker", "o")
    shape_cycle = itertools.cycle(toSequence(shape_cycle))
    color_cycle = scatter_kwargs.get(
        "color", plt.rcParams["axes.prop_cycle"].by_key()["color"]
    )
    color_cycle = itertools.cycle(toSequence(color_cycle))
    if history:
        for i in flags_hist.columns:
            # catch empty but existing history case (flags_meta={})
            if len(flags_meta[i]) == 0:
                continue
            label = (
                flags_meta[i]["kwargs"].get("label", None)
                or flags_meta[i]["func"].split(".")[-1]
            )
            scatter_kwargs.update({"label": label})
            flags_i = flags_hist[i].astype(float)
            if history == "complete":
                scatter_kwargs.update(
                    {"color": next(color_cycle), "marker": next(shape_cycle)}
                )
                _plotFlags(ax, datser, flags_i, na_mask, level, scatter_kwargs)
            if history == "valid":
                # only plot those flags, that do not get altered later on:
                mask = flags_i.eq(flags_vals)
                flags_i[~mask] = np.nan
                # Skip plot, if the test did not have no effect on the all over flagging result. This avoids
                # legend overflow
                if ~(flags_i >= level).any():
                    continue

                # Also skip plot, if all flagged values are np.nans (to catch flag missing and masked results mainly)
                temp_i = datser.index.join(flags_i.index, how="inner")
                if datser[temp_i][flags_i[temp_i].notna()].isna().all() or (
                    "flagMissing" in flags_meta[i]["func"]
                ):
                    continue

                scatter_kwargs.update(
                    {"color": next(color_cycle), "marker": next(shape_cycle)}
                )
                _plotFlags(
                    ax,
                    datser,
                    flags_i,
                    na_mask,
                    level,
                    scatter_kwargs,
                )
        ax.legend()
    else:
        scatter_kwargs.update({"color": next(color_cycle), "marker": next(shape_cycle)})
        _plotFlags(ax, datser, flags_vals, na_mask, level, scatter_kwargs)


def _plotFlags(ax, datser, flags, na_mask, level, scatter_kwargs):
    is_flagged = flags.astype(float) >= level
    is_flagged = is_flagged[~na_mask]
    is_flagged = datser[is_flagged[is_flagged].index]
    ax.scatter(is_flagged.index, is_flagged.values, **scatter_kwargs)


def _insertBlockingNaNs(d, max_gap):
    i = d.index
    gaps = d.reindex(
        pd.date_range(i[0].floor(max_gap), i[-1].ceil(max_gap), freq=max_gap),
        method="bfill",
        tolerance=max_gap,
    )
    gaps = gaps[gaps.isna()]
    return d.reindex(d.index.join(gaps.index, how="outer"))



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import ast

from saqc.constants import ENVIRONMENT
from saqc.core.register import FUNC_MAP


class ConfigExpressionParser(ast.NodeVisitor):
    """
    Generic configuration functions will be rewritten as lambda functions
    and all defined variables will act as arguments, e.g.:
    ``flagGeneric(func=(x != 4) & (y < 3))`` will be rewritten to
    ``lambda x, y: (x != 4) & (y < 3)``

    The main purpose of this class is to identify all variables used in
    a given generic function and to check that it does not violate the
    restrictions imposed onto generic functions.
    """

    SUPPORTED = (
        ast.Str,
        ast.Expression,
        ast.UnaryOp,
        ast.BinOp,
        ast.BitOr,
        ast.BitAnd,
        ast.Num,
        ast.Compare,
        ast.Add,
        ast.Sub,
        ast.Mult,
        ast.Div,
        ast.Pow,
        ast.Mod,
        ast.USub,
        ast.Eq,
        ast.NotEq,
        ast.Gt,
        ast.Lt,
        ast.GtE,
        ast.LtE,
        ast.Invert,
        ast.Name,
        ast.Load,
        ast.Call,
    )

    def __init__(self, node):
        self._args = []
        self.visit(node)

    @property
    def args(self):
        return tuple(dict.fromkeys(self._args))

    def visit_Call(self, node):
        # only non-keyword arguments allowed
        # in generic functions
        for n in node.args:
            self.visit(n)

    def visit_Name(self, node):
        # NOTE:
        # anything not in ENVIRONMENT should be an argument of the generic function
        name = node.id
        if name not in ENVIRONMENT:
            self._args.append(name)
        self.generic_visit(node)

    def generic_visit(self, node):
        if not isinstance(node, self.SUPPORTED):
            raise TypeError(f"invalid expression: '{node}'")
        return super().generic_visit(node)


class ConfigFunctionParser(ast.NodeVisitor):

    SUPPORTED_NODES = (
        ast.Call,
        ast.Num,
        ast.Str,
        ast.keyword,
        ast.NameConstant,
        ast.UnaryOp,
        ast.Name,
        ast.Load,
        ast.Expression,
        ast.Subscript,
        ast.Index,
        ast.USub,
        ast.List,
        ast.Attribute,
    )

    def __init__(self):
        self.kwargs = {}

    def parse(self, node):
        func = self.visit_Call(node)
        return func, self.kwargs

    def visit_Call(self, node):

        if not isinstance(node, ast.Call):
            raise TypeError("expected function call")

        if node.args:
            raise TypeError("only keyword arguments are supported")

        if isinstance(node.func, ast.Attribute):
            func_name = f"{node.func.value.id}.{node.func.attr}"  # type: ignore
        else:
            func_name = node.func.id  # type: ignore

        if func_name not in FUNC_MAP:
            raise NameError(f"unknown function '{func_name}'")

        self.generic_visit(node)
        return func_name

    def visit_keyword(self, node):

        key, value = node.arg, node.value
        check_tree = True

        if key == "func":
            visitor = ConfigExpressionParser(value)
            args = ast.arguments(
                posonlyargs=[],
                kwonlyargs=[],
                kw_defaults=[],
                defaults=[],
                args=[ast.arg(arg=a, annotation=None) for a in visitor.args],
                kwarg=None,
                vararg=None,
            )
            value = ast.Lambda(args=args, body=value)
            # NOTE:
            # don't pass the generated functions down
            # to the checks implemented in this class...
            check_tree = False

        vnode = ast.Assign(targets=[ast.Name(id=key, ctx=ast.Store())], value=value)

        # NOTE:
        # in order to get concrete values out of the AST
        # we compile and evaluate every keyword (e.g. max=100)
        # into the dictionary `self.kwargs`
        # -> after all keywords where visited we end up with
        # a dictionary holding all the passed arguments as
        # real python objects
        co = compile(
            ast.fix_missing_locations(ast.Interactive(body=[vnode])),
            "<ast>",
            mode="single",
        )
        # NOTE: only pass a copy to not clutter the ENVIRONMENT
        # try:
        exec(co, {**ENVIRONMENT}, self.kwargs)

        # let's do some more validity checks
        if check_tree:
            self.generic_visit(value)

    def generic_visit(self, node):
        if not isinstance(node, self.SUPPORTED_NODES):
            raise TypeError(f"invalid node: '{node}'")
        return super().generic_visit(node)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Union, Callable

import numpy as np
import pandas as pd

from saqc.constants import UNFLAGGED
import saqc
from saqc.funcs.interpolation import _SUPPORTED_METHODS


class Interpolation:
    def interpolateByRolling(
        self,
        field: str,
        window: Union[str, int],
        func: Callable[[pd.Series], float] = np.median,
        center: bool = True,
        min_periods: int = 0,
        flag: float = UNFLAGGED,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("interpolateByRolling", locals())

    def interpolateInvalid(
        self,
        field: str,
        method: _SUPPORTED_METHODS,
        order: int = 2,
        limit: int = 2,
        downgrade: bool = False,
        flag: float = UNFLAGGED,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("interpolateInvalid", locals())

    def interpolateIndex(
        self,
        field: str,
        freq: str,
        method: _SUPPORTED_METHODS,
        order: int = 2,
        limit: int = 2,
        downgrade: bool = False,
        **kwargs,
    ) -> saqc.SaQC:
        return self._defer("interpolateIndex", locals())



#!/usr/bin/env python
from click.testing import CliRunner
from pathlib import Path


def test__main__py(tmp_path):
    import saqc.__main__

    # if not run from project root
    projpath = Path(saqc.__file__).parents[1]
    args = [
        "--config",
        Path(projpath, "ressources/data/config.csv"),
        "--data",
        Path(projpath, "ressources/data/data.csv"),
        "--outfile",
        Path(tmp_path, "test.csv"),  # the filesystem temp dir
    ]
    runner = CliRunner()
    for scheme in ["float", "positional", "dmp", "simple"]:
        result = runner.invoke(saqc.__main__.main, args + ["--scheme", scheme])
        assert result.exit_code == 0, result.output



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from typing import Any, Tuple, Union
from typing_extensions import Literal
import pandas as pd
from dateutil.parser import ParserError
import numpy as np
from dios import DictOfSeries

from saqc.constants import BAD, UNFLAGGED
from saqc.core import register, Flags
import warnings

from saqc.core.register import flagging


@register(mask=[], demask=[], squeeze=["field"])
def forceFlags(
    data: DictOfSeries, field: str, flags: Flags, flag: float = BAD, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Set whole column to a flag value.

    Parameters
    ----------
    data : DictOfSeries
        data container
    field : str
        columns name that holds the data
    flags : saqc.Flags
        flags object
    flag : float, default BAD
        flag to set
    kwargs : dict
        unused

    Returns
    -------
    data : DictOfSeries
    flags : saqc.Flags

    See Also
    --------
    clearFlags : set whole column to UNFLAGGED
    flagUnflagged : set flag value at all unflagged positions
    """
    flags[:, field] = flag
    return data, flags


@register(mask=[], demask=[], squeeze=["field"])
def clearFlags(
    data: DictOfSeries, field: str, flags: Flags, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Set whole column to UNFLAGGED.

    Parameters
    ----------
    data : DictOfSeries
        data container
    field : str
        columns name that holds the data
    flags : saqc.Flags
        flags object
    kwargs : dict
        unused

    Returns
    -------
    data : DictOfSeries
    flags : saqc.Flags

    Notes
    -----
    This function ignores the ``dfilter`` keyword, because the data is not relevant
    for processing.
    A warning is triggered if the ``flag`` keyword is given, because the flags are
    always set to `UNFLAGGED`.


    See Also
    --------
    forceFlags : set whole column to a flag value
    flagUnflagged : set flag value at all unflagged positions
    """
    # NOTE: do we really need this?
    if "flag" in kwargs:
        kwargs = {**kwargs}  # copy
        flag = kwargs.pop("flag")
        warnings.warn(f"`flag={flag}` is ignored here.")

    return forceFlags(data, field, flags, flag=UNFLAGGED, **kwargs)


@register(mask=[], demask=[], squeeze=["field"])
def flagUnflagged(
    data: DictOfSeries, field: str, flags: Flags, flag: float = BAD, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Function sets a flag at all unflagged positions.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    flag : float, default BAD
        flag value to set
    kwargs : Dict
        unused

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data

    Notes
    -----
    This function ignores the ``dfilter`` keyword, because the data is not relevant
    for processing.

    See Also
    --------
    clearFlags : set whole column to UNFLAGGED
    forceFlags : set whole column to a flag value
    """
    unflagged = flags[field].isna() | (flags[field] == UNFLAGGED)
    flags[unflagged, field] = flag
    return data, flags


@register(mask=["field"], demask=["field"], squeeze=["field"])
def flagManual(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    mdata: Union[pd.Series, pd.DataFrame, DictOfSeries, list, np.array],
    method: Literal[
        "left-open", "right-open", "closed", "plain", "ontime"
    ] = "left-open",
    mformat: Literal["start-end", "mflag"] = "start-end",
    mflag: Any = 1,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flag data by given, "manually generated" data.

    The data is flagged at locations where `mdata` is equal to a provided flag (`mflag`).
    The format of mdata can be an indexed object, like pd.Series, pd.Dataframe or dios.DictOfSeries,
    but also can be a plain list- or array-like.
    How indexed mdata is aligned to data is specified via the `method` parameter.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    mdata : pd.Series, pd.DataFrame, DictOfSeries, str, list or np.ndarray
        The Data determining, wich intervals are to be flagged, or a string, denoting under which field the data is
        accessable.
    method : {'plain', 'ontime', 'left-open', 'right-open', 'closed'}, default 'plain'
        Defines how mdata is projected on data. Except for the 'plain' method, the methods assume mdata to have an
        index.

        * 'plain': mdata must have the same length as data and is projected one-to-one on data.
        * 'ontime': works only with indexed mdata. mdata entries are matched with data entries that have the same index.
        * 'right-open': mdata defines intervals, values are to be projected on.
          The intervals are defined,

          (1) Either, by any two consecutive timestamps t_1 and 1_2 where t_1 is valued with mflag, or by a series,
          (2) Or, a Series, where the index contains in the t1 timestamps nd the values the respective t2 stamps.

          The value at t_1 gets projected onto all data timestamps t with t_1 <= t < t_2.

        * 'left-open': like 'right-open', but the projected interval now covers all t with t_1 < t <= t_2.
        * 'closed': like 'right-open', but the projected interval now covers all t with t_1 <= t <= t_2.

    mformat : {"start-end", "mflag"}, default "start-end"

        * "start-end": mdata is a Series, where every entry indicates an interval to-flag. The index defines the left
          bound, the value defines the right bound.
        * "mflag": mdata is an array like, with entries containing 'mflag',where flags shall be set. See documentation
          for examples.

    mflag : scalar
        The flag that indicates data points in `mdata`, of wich the projection in data should be flagged.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : original data
    flags : modified flags

    Examples
    --------
    An example for mdata

    >>> mdata = pd.Series([1,0,1], index=pd.to_datetime(['2000-02', '2000-03', '2001-05']))
    >>> mdata
    2000-02-01    1
    2000-03-01    0
    2001-05-01    1
    dtype: int64

    On *dayly* data, with the 'ontime' method, only the provided timestamnps are used.
    Bear in mind that only exact timestamps apply, any offset will result in ignoring
    the timestamp.

    >>> _, fl = flagManual(data, field, flags, mdata, mflag=1, method='ontime')
    >>> fl[field] > UNFLAGGED
    2000-01-31    False
    2000-02-01    True
    2000-02-02    False
    2000-02-03    False
    ..            ..
    2000-02-29    False
    2000-03-01    True
    2000-03-02    False
    Freq: D, dtype: bool

    With the 'right-open' method, the mdata is forward fill:

    >>> _, fl = flagManual(data, field, flags, mdata, mflag=1, method='right-open')
    >>> fl[field] > UNFLAGGED
    2000-01-31    False
    2000-02-01    True
    2000-02-02    True
    ..            ..
    2000-02-29    True
    2000-03-01    False
    2000-03-02    False
    Freq: D, dtype: bool

    With the 'left-open' method, backward filling is used:

    >>> _, fl = flagManual(data, field, flags, mdata, mflag=1, method='left-open')
    >>> fl[field] > UNFLAGGED
    2000-01-31    False
    2000-02-01    False
    2000-02-02    True
    ..            ..
    2000-02-29    True
    2000-03-01    True
    2000-03-02    False
    Freq: D, dtype: bool
    """
    dat = data[field]
    # internal not-mflag-value -> cant go for np.nan
    not_mflag = -1 if mflag == 0 else 0
    if isinstance(mdata, str):
        mdata = data[mdata]

    if isinstance(mdata, (pd.DataFrame, DictOfSeries)):
        mdata = mdata[field]

    hasindex = isinstance(mdata, (pd.Series, pd.DataFrame, DictOfSeries))
    if not hasindex:
        if method != "plain":
            raise ValueError("mdata has no index")
        else:
            mdata = pd.Series(mdata, index=dat.index)

    # check, if intervals where passed in format (index:start-time, data:end-time)
    if mformat == "start-end":
        if method in ["plain", "ontime"]:
            raise ValueError(
                "'Start-End' formatting not compatible to 'plain' or 'ontime' methods"
            )
        else:
            mdata = pd.Series(
                not_mflag,
                index=mdata.index.join(pd.DatetimeIndex(mdata.values), how="outer"),
            )
            mdata[::2] = mflag

    # get rid of values that are neither mflag nor not_mflag (for bw-compatibillity mainly)
    mdata[mdata != mflag] = not_mflag

    # evaluate methods
    if method == "plain":
        pass
    # reindex will do the job later
    elif method == "ontime":
        pass

    elif method in ["left-open", "right-open", "closed"]:
        mdata = mdata.drop(mdata.index[mdata.diff() == 0])
        app_entry = pd.Series(mdata[-1], dat.index.shift(freq="1min")[-1:])
        mdata = mdata.reindex(dat.index.union(mdata.index))

        if method == "right-open":
            mdata = mdata.ffill()

        if method == "left-open":
            mdata = (
                mdata.replace({mflag: not_mflag, not_mflag: mflag})
                .append(app_entry)
                .bfill()
            )

        if method == "closed":
            mdata[mdata.ffill() == mflag] = mflag
            mdata.replace({not_mflag: mflag}, inplace=True)
    else:
        raise ValueError(method)

    mask = mdata == mflag
    mask = mask.reindex(dat.index).fillna(False)

    flags[mask, field] = flag
    return data, flags


@flagging()
def flagDummy(
    data: DictOfSeries, field: str, flags: Flags, **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Function does nothing but returning data and flags.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
    """
    return data, flags



#! /usr/bin/env python
# -*- coding: utf-8 -*-

"""
The module gathers all kinds of timeseries tranformations.
"""
import re
import warnings
from typing import Union
import sys

import pandas as pd
import numpy as np
import numba as nb
from sklearn.neighbors import NearestNeighbors
from scipy.stats import iqr, median_abs_deviation
from scipy.signal import filtfilt, butter
import numpy.polynomial.polynomial as poly


def identity(ts):
    # identity function
    return ts


def count(ts):
    # count is a dummy to trigger according built in count method of resamplers when
    # passed to aggregate2freq. For consistency reasons, it works accordingly when
    # applied directly:
    return ts.count()


def first(ts):
    # first is a dummy to trigger according built in count method of resamplers when
    # passed to aggregate2freq. For consistency reasons, it works accordingly when
    # applied directly:
    return ts.first()


def last(ts):
    # last is a dummy to trigger according built in count method of resamplers when
    # passed to aggregate2freq. For consistency reasons, it works accordingly when
    # applied directly:
    return ts.last()


def zeroLog(ts):
    # zero log returns np.nan instead of -np.inf, when passed 0. Usefull, because
    # in internal processing, you only have to check for nan values if you need to
    # remove "invalidish" values from the data.
    log_ts = np.log(ts)
    log_ts[log_ts == -np.inf] = sys.float_info.min
    return log_ts


def derivative(ts, unit="1min"):
    # calculates derivative of timeseries, expressed in slope per "unit"
    return ts / (deltaT(ts, unit=unit))


def deltaT(ts, unit="1min"):
    # calculates series of time gaps in ts
    return (
        ts.index.to_series().diff().dt.total_seconds()
        / pd.Timedelta(unit).total_seconds()
    )


def difference(ts):
    # NOTE: index of input series gets lost!
    return np.diff(ts, prepend=np.nan)


def rateOfChange(ts):
    return difference(ts) / ts


def relativeDifference(ts):
    res = ts - 0.5 * (np.roll(ts, +1) + np.roll(ts, -1))
    res[0] = np.nan
    res[-1] = np.nan
    return res


def scale(ts, target_range=1, projection_point=None):
    # scales input series to have values ranging from - target_rang to + target_range
    if not projection_point:
        projection_point = np.max(np.abs(ts))
    return (ts / projection_point) * target_range


def normScale(ts):
    # func scales series to [0,1] interval and projects constant series onto 0.5
    ts_min = ts.min()
    ts_max = ts.max()
    if ts_min == ts_max:
        return pd.Series(data=0.5, index=ts.index)
    else:
        return (ts - ts_min) / (ts.max() - ts_min)


def standardizeByMean(ts):
    # standardization with mean and probe variance
    return (ts - np.mean(ts)) / np.std(ts, ddof=1)


def standardizeByMedian(ts):
    # standardization with median (MAD)
    # NO SCALING
    return (ts - np.median(ts)) / median_abs_deviation(ts, nan_policy="omit")


def standardizeByIQR(ts):
    # standardization with median and inter quantile range
    return (ts - np.median(ts)) / iqr(ts, nan_policy="omit")


def kNN(in_arr, n_neighbors, algorithm="ball_tree", metric="minkowski", p=2):
    # k-nearest-neighbor search

    nbrs = NearestNeighbors(
        n_neighbors=n_neighbors, algorithm=algorithm, metric=metric, p=p
    ).fit(in_arr.reshape(in_arr.shape[0], -1))
    return nbrs.kneighbors()


def maxGap(in_arr):
    """
    Search for the maximum gap in an array of sorted distances (func for scoring kNN
    distance matrice)
    """
    return max(in_arr[0], max(np.diff(in_arr)))


@nb.njit
def _maxConsecutiveNan(arr, max_consec):
    # checks if arr (boolean array) has not more then "max_consec" consecutive True
    # values
    current = 0
    idx = 0
    while idx < arr.size:
        while idx < arr.size and arr[idx]:
            current += 1
            idx += 1
        if current > max_consec:
            return False
        current = 0
        idx += 1
    return True


def validationTrafo(data, max_nan_total, max_nan_consec):
    # data has to be boolean. False=Valid Value, True=invalid Value function returns
    # True-array of input array size for invalid input arrays False array for valid
    # ones
    data = data.copy()
    if (max_nan_total is np.inf) & (max_nan_consec is np.inf):
        return data

    # nan_mask = np.isnan(data)

    if data.sum() <= max_nan_total:
        if max_nan_consec is np.inf:
            data[:] = False
            return data
        elif _maxConsecutiveNan(np.asarray(data), max_nan_consec):
            data[:] = False
        else:
            data[:] = True
    else:
        data[:] = True

    return data


def stdQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    return np.nanstd(
        data[~validationTrafo(data.isna(), max_nan_total, max_nan_consec)], ddof=1
    )


def varQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    return np.nanvar(
        data[~validationTrafo(data.isna(), max_nan_total, max_nan_consec)], ddof=1
    )


def meanQC(data, max_nan_total=np.inf, max_nan_consec=np.inf):
    return np.nanmean(
        data[~validationTrafo(data.isna(), max_nan_total, max_nan_consec)]
    )


def interpolateNANs(
    data, method, order=2, inter_limit=2, downgrade_interpolation=False
):
    """
    The function interpolates nan-values (and nan-grids) in timeseries data. It can
    be passed all the method keywords from the pd.Series.interpolate method and will
    than apply this very methods. Note, that the limit keyword really restricts
    the interpolation to chunks, not containing more than "limit" nan entries (
    thereby not being identical to the "limit" keyword of pd.Series.interpolate).

    :param data:                    pd.Series or np.array. The data series to be interpolated
    :param method:                  String. Method keyword designating interpolation method to use.
    :param order:                   Integer. If your desired interpolation method needs an order to be passed -
                                    here you pass it.
    :param inter_limit:             Integer. Default = 2. Limit up to which consecutive nan - values in the data get
                                    replaced by interpolation.
                                    Its default value suits an interpolation that only will apply to points of an
                                    inserted frequency grid. (regularization by interpolation)
                                    Gaps wider than "limit" will NOT be interpolated at all.
    :param downgrade_interpolation:  Boolean. Default False. If True:
                                    If a data chunk not contains enough values for interpolation of the order "order",
                                    the highest order possible will be selected for that chunks interpolation.

    :return:
    """
    inter_limit = int(inter_limit)
    data = pd.Series(data, copy=True)
    gap_mask = data.isna().rolling(inter_limit, min_periods=0).sum() != inter_limit

    if inter_limit == 2:
        gap_mask = gap_mask & gap_mask.shift(-1, fill_value=True)
    else:
        gap_mask = (
            gap_mask.replace(True, np.nan)
            .fillna(method="bfill", limit=inter_limit)
            .replace(np.nan, True)
            .astype(bool)
        )

    pre_index = data.index
    data = data[gap_mask]

    if method in ["linear", "time"]:

        data.interpolate(
            method=method, inplace=True, limit=inter_limit - 1, limit_area="inside"
        )

    else:
        dat_name = data.name
        gap_mask = (~gap_mask).cumsum()
        data = pd.merge(gap_mask, data, how="inner", left_index=True, right_index=True)

        def _interpolWrapper(x, wrap_order=order, wrap_method=method):
            if x.count() > wrap_order:
                try:
                    return x.interpolate(method=wrap_method, order=int(wrap_order))
                except (NotImplementedError, ValueError):
                    warnings.warn(
                        f"Interpolation with method {method} is not supported at order "
                        f"{wrap_order}. and will be performed at order {wrap_order-1}"
                    )
                    return _interpolWrapper(x, int(wrap_order - 1), wrap_method)
            elif x.size < 3:
                return x
            else:
                if downgrade_interpolation:
                    return _interpolWrapper(x, int(x.count() - 1), wrap_method)
                else:
                    return x

        data = data.groupby(data.columns[0]).transform(_interpolWrapper)
        # squeezing the 1-dimensional frame resulting from groupby for consistency
        # reasons
        data = data.squeeze(axis=1)
        data.name = dat_name
    data = data.reindex(pre_index)

    return data


def aggregate2Freq(
    data: pd.Series,
    method,
    freq,
    agg_func,
    fill_value=np.nan,
    max_invalid_total=None,
    max_invalid_consec=None,
):
    """
    The function aggregates values to an equidistant frequency grid with func.
    Timestamps that gets no values projected, get filled with the fill-value. It
    also serves as a replacement for "invalid" intervals.
    """
    methods = {
        # offset, closed, label
        "nagg": lambda f: (f / 2, "left", "left"),
        "bagg": lambda _: (pd.Timedelta(0), "left", "left"),
        "fagg": lambda _: (pd.Timedelta(0), "right", "right"),
    }

    # filter data for invalid patterns (since filtering is expensive we pre-check if
    # it is demanded)
    if max_invalid_total is not None or max_invalid_consec is not None:
        if pd.isna(fill_value):
            temp_mask = data.isna()
        else:
            temp_mask = data == fill_value

        temp_mask = temp_mask.groupby(pd.Grouper(freq=freq)).transform(
            validationTrafo,
            max_nan_total=max_invalid_total,
            max_nan_consec=max_invalid_consec,
        )
        data[temp_mask] = fill_value

    freq = pd.Timedelta(freq)
    offset, closed, label = methods[method](freq)

    resampler = data.resample(
        freq, closed=closed, label=label, origin="start_day", offset=offset
    )

    # count valid values
    counts = resampler.count()

    # native methods of resampling (median, mean, sum, ..) are much faster than apply
    try:
        check_name = re.sub("^nan", "", agg_func.__name__)
        # a nasty special case: if function "count" was passed, we not want empty
        # intervals to be replaced by fill_value:
        if check_name == "count":
            data = counts.copy()
            counts[:] = np.nan
        else:
            data = getattr(resampler, check_name)()
    except AttributeError:
        data = resampler.apply(agg_func)

    # we custom fill bins that have no value
    data[counts == 0] = fill_value

    # undo the temporary shift, to mimic centering the frequency
    if method == "nagg":
        data.index += offset

    return data


def shift2Freq(
    data: Union[pd.Series, pd.DataFrame], method: str, freq: str, fill_value
):
    """
    shift timestamps backwards/forwards in order to align them with an equidistant
    frequency grid. Resulting Nan's are replaced with the fill-value.
    """

    methods = {
        "fshift": lambda freq: ("ffill", pd.Timedelta(freq)),
        "bshift": lambda freq: ("bfill", pd.Timedelta(freq)),
        "nshift": lambda freq: ("nearest", pd.Timedelta(freq) / 2),
    }
    direction, tolerance = methods[method](freq)
    target_ind = pd.date_range(
        start=data.index[0].floor(freq),
        end=data.index[-1].ceil(freq),
        freq=freq,
        name=data.index.name,
    )
    return data.reindex(
        target_ind, method=direction, tolerance=tolerance, fill_value=fill_value
    )


def butterFilter(
    x, cutoff, nyq=0.5, filter_order=2, fill_method="linear", filter_type="low"
):
    """
    Applies butterworth filter.
    `x` is expected to be regularly sampled.

    Parameters
    ----------
    x: pd.Series
        input timeseries

    cutoff: float
        The cutoff-frequency, expressed in multiples of the sampling rate.

    nyq: float
        The niquist-frequency. expressed in multiples if the sampling rate.

    fill_method: Literal[‘nearest’, ‘zero’, ‘slinear’, ‘quadratic’, ‘cubic’, ‘spline’, ‘barycentric’, ‘polynomial’]
        Fill method to be applied on the data before filtering (butterfilter cant
        handle ''np.nan''). See documentation of pandas.Series.interpolate method for
        details on the methods associated with the different keywords.


    Returns
    -------
    """
    na_mask = x.isna()
    x = x.interpolate(fill_method).interpolate("ffill").interpolate("bfill")
    b, a = butter(N=filter_order, Wn=cutoff / nyq, btype=filter_type)
    y = pd.Series(filtfilt(b, a, x), x.index, name=x.name)
    y[na_mask] = np.nan
    return y


@nb.njit
def _coeffMat(x, deg):
    # helper function to construct numba-compatible polynomial fit function
    mat_ = np.zeros(shape=(x.shape[0], deg + 1))
    const = np.ones_like(x)
    mat_[:, 0] = const
    mat_[:, 1] = x
    if deg > 1:
        for n in range(2, deg + 1):
            mat_[:, n] = x ** n
    return mat_


@nb.jit
def _fitX(a, b):
    # helper function to construct numba-compatible polynomial fit function
    # linalg solves ax = b
    det_ = np.linalg.lstsq(a, b)[0]
    return det_


@nb.jit
def _fitPoly(x, y, deg):
    # a numba compatible polynomial fit function
    a = _coeffMat(x, deg)
    p = _fitX(a, y)
    # Reverse order so p[0] is coefficient of highest order
    return p[::-1]


@nb.jit
def evalPolynomial(P, x):
    # a numba compatible polynomial evaluator
    result = 0
    for coeff in P:
        result = x * result + coeff
    return result


def polyRollerNumba(in_slice, miss_marker, val_range, center_index, poly_deg):
    # numba compatible function to roll with when modelling data with polynomial model
    miss_mask = in_slice == miss_marker
    x_data = val_range[~miss_mask]
    y_data = in_slice[~miss_mask]
    fitted = _fitPoly(x_data, y_data, deg=poly_deg)
    return evalPolynomial(fitted, center_index)


def polyRollerNoMissingNumba(in_slice, val_range, center_index, poly_deg):
    # numba compatible function to roll with when modelling data with polynomial model -
    # it is assumed, that in slice is an equidistant sample
    fitted = _fitPoly(val_range, in_slice, deg=poly_deg)
    return evalPolynomial(fitted, center_index)


def polyRoller(in_slice, miss_marker, val_range, center_index, poly_deg):
    # function to roll with when modelling data with polynomial model
    miss_mask = in_slice == miss_marker
    x_data = val_range[~miss_mask]
    y_data = in_slice[~miss_mask]
    fitted = poly.polyfit(x=x_data, y=y_data, deg=poly_deg)
    return poly.polyval(center_index, fitted)


def polyRollerNoMissing(in_slice, val_range, center_index, poly_deg):
    # function to roll with when modelling data with polynomial model
    # it is assumed, that in slice is an equidistant sample
    fitted = poly.polyfit(x=val_range, y=in_slice, deg=poly_deg)
    return poly.polyval(center_index, fitted)


def polyRollerIrregular(in_slice, center_index_ser, poly_deg):
    # a function to roll with, for polynomial fitting of data not having an
    # equidistant frequency grid. (expects to get passed pandas timeseries),
    # so raw parameter of rolling.apply should be set to False.
    x_data = ((in_slice.index - in_slice.index[0]).total_seconds()) / 60
    fitted = poly.polyfit(x_data, in_slice.values, poly_deg)
    center_pos = int(len(in_slice) - center_index_ser[in_slice.index[-1]])
    return poly.polyval(x_data[center_pos], fitted)


def expModelFunc(x, a=0, b=0, c=0):
    # exponential model function, used in optimization contexts (drift correction)
    return a + b * (np.exp(c * x) - 1)


def expDriftModel(x, c, origin, target):
    c = abs(c)
    b = (target - origin) / (np.exp(c) - 1)
    return expModelFunc(x, origin, b, c)


def linearDriftModel(x, origin, target):
    return origin + x * target


def linearInterpolation(data, inter_limit=2):
    return interpolateNANs(data, "time", inter_limit=inter_limit)


def polynomialInterpolation(data, inter_limit=2, inter_order=2):
    return interpolateNANs(
        data, "polynomial", inter_limit=inter_limit, order=inter_order
    )



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="saqc",
    version="2.0.0",
    author="Bert Palm, David Schaefer, Peter Luenenschloss, Lennard Schmidt",
    author_email="david.schaefer@ufz.de",
    description="Data quality checking and processing tool/framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.ufz.de/rdm-software/saqc",
    packages=find_packages(exclude=("tests",)),
    python_requires=">=3.7, <3.10",
    install_requires=[
        "numpy==1.20.*",
        "pandas==1.3.*",
        "scipy==1.7.*",
        "scikit-learn==1.0.*",
        "numba==0.54.*",
        "matplotlib==3.4.*",
        "Click==8.0.*",
        "pyarrow==4.0.*",
        "typing_extensions==3.10.*",
        "outlier-utils==0.0.3",
        "dtw==1.4.*",
        "seaborn==0.11.*",
    ],
    license_files=("LICENSE.md",),
    entry_points={
        "console_scripts": ["saqc=saqc.__main__:main"],
    },
)



import pytest

from saqc.lib.ts_operators import butterFilter
import pandas as pd


def test_butterFilter():
    assert (
        butterFilter(pd.Series([1, -1] * 100), cutoff=0.1) - pd.Series([1, -1] * 100)
    ).mean() < 0.5



#!/usr/bin/env python
from __future__ import annotations

import inspect
import warnings
from typing import Dict, Tuple, Callable, Sequence, Any
import functools
import numpy as np
import pandas as pd
import dios

from saqc.constants import UNFLAGGED, FILTER_ALL
from saqc.core.flags import Flags, History
from saqc.lib.tools import squeezeSequence, toSequence

# NOTE:
# the global SaQC function store,
# will be filled by calls to register
FUNC_MAP: Dict[str, Callable] = {}

_is_list_like = pd.api.types.is_list_like


class FunctionWrapper:
    def __init__(
        self,
        func: Callable,
        mask: list,
        demask: list,
        squeeze: list,
        multivariate: bool = False,
        handles_target: bool = False,
    ):
        # todo:
        #  - meta only is written with squeeze

        self.func = func
        self.func_name = func.__name__
        self.func_signature = inspect.signature(func)

        # ensure type and all elements exist in signature
        self._checkDecoratorKeywords(mask, demask, squeeze)
        if multivariate and not handles_target:
            raise ValueError("multivariate=True requires handle_target=True")

        self.decorator_mask = mask
        self.decorator_demask = demask
        self.decorator_squeeze = squeeze
        self.multivariate = multivariate
        self.handles_target = handles_target

        # set in __call__
        self.data = None
        self.flags = None
        self.fields = None
        self.args = None
        self.kwargs = None
        self.mask_thresh = None
        self.stored_data = None

        # make ourself look like the wrapped function, especially the docstring
        functools.update_wrapper(self, func)

    def _checkDecoratorKeywords(self, mask, demask, squeeze):
        params = self.func_signature.parameters.keys()
        for dec_arg, name in zip(
            [mask, demask, squeeze], ["mask", "demask", "squeeze"]
        ):
            typeerr = TypeError(
                f"type of decorator argument '{name}' must "
                f"be a list of strings, not {repr(type(dec_arg))}"
            )
            if not isinstance(dec_arg, list):
                raise typeerr
            for elem in dec_arg:
                if not isinstance(elem, str):
                    raise typeerr
                if elem not in params:
                    raise ValueError(
                        f"passed value {repr(elem)} in {repr(name)} is not an "
                        f"parameter in decorated function {repr(self.func_name)}"
                    )

    @staticmethod
    def _argnamesToColumns(names: list, values: dict):
        clist = []
        for name in names:
            value = values[name]  # eg. the value behind 'field'

            # NOTE: do not change order of the tests
            if value is None:
                pass
            elif isinstance(value, str):
                clist.append(value)
            # we ignore DataFrame, Series, DictOfSeries
            # and high order types alike
            elif hasattr(value, "columns"):
                pass
            elif _is_list_like(value) and all([isinstance(e, str) for e in value]):
                clist += value
        return pd.Index(clist)

    @staticmethod
    def _warn(missing, source):
        if len(missing) == 0:
            return
        action = source + "ed"
        obj = "flags" if source == "squeeze" else "data"
        warnings.warn(
            f"Column(s) {repr(missing)} cannot not be {action} "
            f"because they are not present in {obj}. ",
            RuntimeWarning,
        )

    def __call__(
        self, data: dios.DictOfSeries, field: str, flags: Flags, *args, **kwargs
    ) -> Tuple[dios.DictOfSeries, Flags]:
        """
        This wraps a call to a saqc function.

        Before the saqc function call it copies flags and maybe mask data (inplace).
        After the call it maybe squeezes modified histories and maybe reinsert the
        masked data locations.

        If the squeezing and/or the masking and/or the demasking will happen, depends
        on the decorator keywords `handles` and `datamask`. See ``_determineActions``,
        for that.
        """
        # keep this the original values
        self.data = data
        self.flags = flags
        self.fields = toSequence(field)
        self.args = args
        self.kwargs = self._checkKwargs(kwargs)

        self.mask_thresh = self._getMaskingThresh()

        # skip (data, field, flags)
        names = list(self.func_signature.parameters.keys())[3 : 3 + len(args)]
        all_args = {"field": field, **dict(zip(names, args)), **kwargs}

        # find columns that need masking
        columns = self._argnamesToColumns(self.decorator_mask, all_args)
        self._warn(columns.difference(self.data.columns).to_list(), source="mask")
        columns = columns.intersection(self.data.columns)

        masked, stored = self._maskData(
            data=self.data,
            flags=self.flags,
            columns=columns,
            thresh=self.mask_thresh,
        )
        self.data = masked
        self.stored_data = stored

        args, kwargs = self._prepareArgs()
        data, flags = self.func(*args, **kwargs)

        # find columns that need squeezing
        columns = self._argnamesToColumns(self.decorator_squeeze, all_args)
        self._warn(columns.difference(flags.columns).to_list(), source="squeeze")
        columns = columns.intersection(flags.columns)

        # if the function did not want to set any flags at all,
        # we assume a processing function that altered the flags
        # in an unpredictable manner or do nothing with the flags.
        # in either case we take the returned flags as the new truth.
        if columns.empty:
            result_flags = flags
        else:
            # even if this looks like a noop for columns=[],
            # it returns the old instead the new flags and
            # therefore ignores any possible processing changes
            result_flags = self._squeezeFlags(flags, columns)

        # find columns that need demasking
        columns = self._argnamesToColumns(self.decorator_demask, all_args)
        self._warn(columns.difference(data.columns).to_list(), source="demask")
        columns = columns.intersection(data.columns)

        result_data = self._unmaskData(data, self.stored_data, columns=columns)

        return result_data, result_flags

    @staticmethod
    def _checkKwargs(kwargs: dict) -> dict[str, Any]:
        if "dfilter" in kwargs and not isinstance(
            kwargs["dfilter"], (bool, float, int)
        ):
            raise TypeError(f"'dfilter' must be of type bool or float")
        return kwargs

    def _prepareArgs(self) -> Tuple[tuple, dict[str, Any]]:
        """
        Prepare the args and kwargs passed to the function
        Returns
        -------
        args: tuple
            arguments to be passed to the actual call
        kwargs: dict
            keyword-arguments to be passed to the actual call
        """
        kwargs = self.kwargs.copy()
        kwargs["dfilter"] = self.mask_thresh

        # always pass a list to multivariate functions and
        # unpack single element lists for univariate functions
        if self.multivariate:
            field = self.fields
        else:
            field = squeezeSequence(self.fields)

        args = self.data, field, self.flags.copy(), *self.args
        return args, kwargs

    def _getMaskingThresh(self) -> float:
        """
        Generate a float threshold by the value of the `dfilter` keyword

        Returns
        -------
        threshold: float
            All data gets masked, if the flags are equal or worse than the threshold.

        Notes
        -----
        If ``dfilter`` is **not** in the kwargs, the threshold defaults to `FILTER_ALL`.
        For any floatish value, it is taken as the threshold.
        """
        if "dfilter" not in self.kwargs:
            return FILTER_ALL
        return float(self.kwargs["dfilter"])  # handle int

    def _createMeta(self) -> dict:
        return {
            "func": self.func_name,
            "args": self.args,
            "kwargs": self.kwargs,
        }

    def _squeezeFlags(self, flags: Flags, columns: pd.Index) -> Flags:
        """
        Generate flags from the temporary result-flags and the original flags.

        Parameters
        ----------
        flags : Flags
            The flags-frame, which is the result from a saqc-function

        Returns
        -------
        Flags
        """
        out = self.flags.copy()  # the old flags
        meta = self._createMeta()
        for col in columns:

            # todo: shouldn't we fail or warn here or even have a explicit test upstream
            #  because the function should ensure consistence, especially because
            #  a empty history maybe issnt what is expected, but this happens silently
            if col not in out:  # ensure existence
                out.history[col] = History(index=flags.history[col].index)

            old_history = out.history[col]
            new_history = flags.history[col]

            # We only want to add new columns, that were appended during the last
            # function call. If no such columns exist, we end up with an empty
            # new_history.
            start = len(old_history.columns)
            new_history = self._sliceHistory(new_history, slice(start, None))

            squeezed = new_history.max(raw=True)
            out.history[col] = out.history[col].append(squeezed, meta=meta)

        return out

    @staticmethod
    def _sliceHistory(history: History, sl: slice) -> History:
        history.hist = history.hist.iloc[:, sl]
        history.meta = history.meta[sl]
        return history

    @staticmethod
    def _maskData(
        data: dios.DictOfSeries, flags: Flags, columns: Sequence[str], thresh: float
    ) -> Tuple[dios.DictOfSeries, dios.DictOfSeries]:
        """
        Mask data with Nans, if the flags are worse than a threshold.
            - mask only passed `columns` (preselected by `datamask`-kw from decorator)

        Returns
        -------
        masked : dios.DictOfSeries
            masked data, same dim as original
        mask : dios.DictOfSeries
            dios holding iloc-data-pairs for every column in `data`
        """
        mask = dios.DictOfSeries(columns=columns)

        # we use numpy here because it is faster
        for c in columns:
            col_mask = _isflagged(flags[c].to_numpy(), thresh)

            if col_mask.any():
                col_data = data[c].to_numpy(dtype=np.float64)

                mask[c] = pd.Series(col_data[col_mask], index=np.where(col_mask)[0])

                col_data[col_mask] = np.nan
                data[c] = col_data

        return data, mask

    @staticmethod
    def _unmaskData(
        data: dios.DictOfSeries, mask: dios.DictOfSeries, columns: pd.Index = None
    ) -> dios.DictOfSeries:
        """
        Restore the masked data.

        Notes
        -----
        - Even if this returns data, it works inplace !
        - `mask` is not a boolean mask, instead it holds the original values.
          The index of mask is numeric and represent the integer location
          in the original data.
        """
        if columns is None:
            columns = data.columns  # field was in old, is in mask and is in new
        columns = mask.columns.intersection(columns)

        for c in columns:

            # ignore
            if data[c].empty or mask[c].empty:
                continue

            # get the positions of values to unmask
            candidates = mask[c]
            # if the mask was removed during the function call, don't replace
            unmask = candidates[data[c].iloc[candidates.index].isna().to_numpy()]
            if unmask.empty:
                continue
            data[c].iloc[unmask.index] = unmask

        return data


def register(
    mask: list[str],
    demask: list[str],
    squeeze: list[str],
    multivariate: bool = False,
    handles_target: bool = False,
):
    """
    Generalized decorator for any saqc functions.

    Before the call of the decorated function:
    - data gets masked by flags according to `dfilter`

    After the call of the decorated function:
    - data gets demasked (original data is written back)
    - flags gets squeezed (only one history column append per call)

    Parameters
    ----------
    mask : list of string
        A list of all parameter of the decorated function, that specify a column in
        data, that is read by the function and therefore should be masked by flags.

        The masking takes place before the call of the decorated function and
        temporary sets data to `NaN` at flagged locations. It is undone by ``demask``.
        The threshold of which data is considered to be flagged can be controlled
        via ``dfilter``, a parameter each function takes.

    demask : list of string
        A list of all parameter of the decorated function, that specify a column in
        data, that was masked (see ``mask``) and needs unmasking after the call.

        The unmasking replace all remaining(!) ``NaN`` by its original values from
        before the call of the decorated function.

    squeeze : list of string
        A list of all parameter of the decorated function, that specify a column in
        flags, that is written by the function.

        The squeezing combines multiple columns in the history of flags to one
        single column. This is because, multiple writes to flags, (eg. using
        ``flags[:,'a'] = 255`` twice) will result in multiple history columns,
        but should considered as a single column, because only one function call
        happened.

    multivariate : bool, default False
        If ``True``, the decorated function, process multiple data or flags
        columns at once. Therefore the decorated function must handle a list
        of columns in the parameter ``field``.

        If ``False``, the decorated function must take a single column (``str``)
        in ``field``.

    handles_target : bool, default False
        If ``True``, the decorated function, handles the target parameter by
        itself. Mandatory for multivariate functions.
    """

    def inner(func):
        wrapper = FunctionWrapper(
            func, mask, demask, squeeze, multivariate, handles_target
        )
        FUNC_MAP[wrapper.func_name] = wrapper
        return wrapper

    return inner


def flagging(**kwargs):
    """
    Default decorator for univariate flagging functions.

    Before the call of the decorated function:
    - `data[field]` gets masked by `flags[field]` according to `dfilter`
    After the call of the decorated function:
    - `data[field]` gets demasked (original data is written back)
    - `flags[field]` gets squeezed (only one history column append per call) if needed

    Notes
    -----
    For full control over masking, demasking and squeezing or to implement
    a multivariate function (multiple in- or outputs) use the `@register` decorator.

    See Also
    --------
        resister: generalization of of this function
    """
    if kwargs:
        raise ValueError("use '@register' to pass keywords")
    return register(mask=["field"], demask=["field"], squeeze=["field"])


def processing(**kwargs):
    """
    Default decorator for univariate processing functions.

    - no masking of data
    - no demasking of data
    - no squeezing of flags

    Notes
    -----
    For full control over masking, demasking and squeezing or to implement
    a multivariate function (multiple in- or outputs) use the `@register` decorator.

    See Also
    --------
        resister: generalization of of this function
    """
    if kwargs:
        raise ValueError("use '@register' to pass keywords")
    return register(mask=[], demask=[], squeeze=[])


def _isflagged(flagscol: np.ndarray | pd.Series, thresh: float) -> np.array | pd.Series:
    """
    Return a mask of flags accordingly to `thresh`. Return type is same as flags.
    """
    if not isinstance(thresh, (float, int)):
        raise TypeError(f"thresh must be of type float, not {repr(type(thresh))}")

    if thresh == FILTER_ALL:
        return flagscol > UNFLAGGED

    return flagscol >= thresh



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from typing import Optional, Callable, Tuple, Union
import numpy as np
import pandas as pd
from dios import DictOfSeries

from saqc.core import register, Flags


@register(mask=["field"], demask=[], squeeze=[])
def transform(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    func: Callable[[pd.Series], pd.Series],
    freq: Optional[Union[float, str]] = None,
    **kwargs
) -> Tuple[DictOfSeries, Flags]:
    """
    Function to transform data columns with a transformation that maps series onto series of the same length.

    Note, that flags get preserved.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-transformed.
    flags : saqc.Flags
        Container to store quality flags to data.
    func : Callable[{pd.Series, np.array}, np.array]
        Function to transform data[field] with.
    freq : {None, float, str}, default None
        Determines the segmentation of the data into partitions, the transformation is applied on individually

        * ``np.inf``: Apply transformation on whole data set at once
        * ``x`` > 0 : Apply transformation on successive data chunks of periods length ``x``
        * Offset String : Apply transformation on successive partitions of temporal extension matching the passed offset
          string

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
        Data values may have changed relatively to the data input.
    flags : saqc.Flags
        The quality flags of data
    """
    val_ser = data[field].copy()
    # partitioning
    if not freq:
        freq = val_ser.shape[0]

    if isinstance(freq, str):
        grouper = pd.Grouper(freq=freq)
    else:
        grouper = pd.Series(data=np.arange(0, val_ser.shape[0]), index=val_ser.index)
        grouper = grouper.transform(lambda x: int(np.floor(x / freq)))

    partitions = val_ser.groupby(grouper)

    for _, partition in partitions:
        if partition.empty:
            continue
        val_ser[partition.index] = func(partition)

    data[field] = val_ser
    return data, flags



from .test_setup import *


def test__len__(datetime_series, maxlen=10):
    dios = DictOfSeries()
    assert len(dios) == 0

    for i in range(maxlen):
        dios[f"c{i}"] = datetime_series.copy()
        assert len(dios) == i + 1

    for i in reversed(range(maxlen)):
        assert len(dios) == i + 1
        del dios[f"c{i}"]

    assert len(dios) == 0



# do not import dios-stuff here
import operator as op


_OP1_MAP = {
    op.inv: "~",
    op.neg: "-",
    op.abs: "abs()",
}

_OP2_COMP_MAP = {
    op.eq: "==",
    op.ne: "!=",
    op.le: "<=",
    op.ge: ">=",
    op.gt: ">",
    op.lt: "<",
}

_OP2_BOOL_MAP = {
    op.and_: "&",
    op.or_: "|",
    op.xor: "^",
}
_OP2_ARITH_MAP = {
    op.add: "+",
    op.sub: "-",
    op.mul: "*",
    op.pow: "**",
}

_OP2_DIV_MAP = {
    op.mod: "%",
    op.truediv: "/",
    op.floordiv: "//",
}

OP_MAP = _OP2_COMP_MAP.copy()
OP_MAP.update(_OP2_BOOL_MAP)
OP_MAP.update(_OP2_ARITH_MAP)
OP_MAP.update(_OP2_DIV_MAP)
OP_MAP.update(_OP1_MAP)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from saqc.constants import BAD
import saqc


class Pattern:
    def flagPatternByDTW(
        self,
        field,
        reference,
        max_distance=0.0,
        normalize=True,
        plot=False,
        flag=BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagPatternByDTW", locals())



import gc
from .generate_testsets import get_random_df_and_dios


def calc_mem(rows, cols, shifted=False, dtypesz=(64 / 8)):
    if shifted:
        idxsz = 8 * rows * cols
        # additional nans are inserted exactly as many as variables
        rowsz = rows * cols * dtypesz
    else:
        idxsz = 8 * rows
        rowsz = rows * dtypesz

    return idxsz + rowsz * cols


def bytes2hread(bytes):
    i = 0
    units = ["B", "kB", "MB", "GB", "TB"]
    while bytes > 1000:
        bytes /= 1024
        i += 1
        if i == 4:
            break
    return bytes, units[i]


def rows_by_time(nsec, mdays):
    """calc the number of values for one value every n seconds in m days
    :param nsec: n seconds a value occur
    :param mdays: this many days of data
    :return: rows thats needed
    """
    return int((60 / nsec) * 60 * 24 * mdays)


if __name__ == "__main__":

    # dios      - linear in rows and colums, same size for r=10,c=100 or r=100,c=10
    do_real_check = True
    cols = 10
    rows = 100000
    # rows = rows_by_time(nsec=600, mdays=365*2)

    mem = calc_mem(rows, cols, shifted=False)
    memsh = calc_mem(rows, cols, shifted=True)

    df, dios = get_random_df_and_dios(rows, cols, disalign=False, randstart=True)
    dios_mem = dios.memory_usage()
    print(f"dios:\n-----------")
    print("mem: ", *bytes2hread(dios_mem))
    print("entries:", sum([len(dios[e]) for e in dios]))
    print()

    ratio = (1 / (memsh - mem)) * dios_mem

    mem = bytes2hread(mem)
    memsh = bytes2hread(memsh)

    print("df - best case\n---------")
    print("mem: ", *mem)
    print("entries:", rows)
    print()
    print("df - worst case\n---------")
    print("mem :", *memsh)
    print("entries:", rows * cols)

    print()
    print(f"dfbest, dios, dfworst: 0%, {round(ratio, 4)*100}%, 100% ")

    if not do_real_check:
        exit(0)

    proveMeRight = False

    if proveMeRight:
        # best case
        print()
        print("best case proove")
        dfb, _ = get_random_df_and_dios(rows, cols, disalign=False, randstart=False)
        dfb.info(memory_usage="deep", verbose=False)

    print()
    print("rand start, same freq")
    df.info(memory_usage="deep", verbose=False)
    print("entries:", sum([len(df[e]) for e in df]))

    print()
    print("rand start, rand freq")
    df, _ = get_random_df_and_dios(rows, cols, disalign="random", randstart=True)
    df.info(memory_usage="deep", verbose=False)
    print("entries:", sum([len(df[e]) for e in df]))

    if proveMeRight:
        # worst case
        print()
        print("worst case proove")
        df, _ = get_random_df_and_dios(rows, cols, disalign=True, randstart=False)
        df.info(memory_usage="deep", verbose=False)

    gc.collect()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from saqc.constants import BAD
import saqc


class Constants:
    def flagByVariance(
        self,
        field: str,
        window: str = "12h",
        thresh: float = 0.0005,
        maxna: int = None,
        maxna_group: int = None,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagByVariance", locals())

    def flagConstants(
        self, field: str, thresh: float, window: str, flag: float = BAD, **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagConstants", locals())



#! /usr/bin/env python
# -*- coding: utf-8 -*-
"""The system for automated quality controll package."""

__version__ = "1.4"

# import order: from small to big
from saqc.constants import (
    UNFLAGGED,
    GOOD,
    DOUBTFUL,
    BAD,
)

from saqc.core import (
    Flags,
    SaQC,
    fromConfig,
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-


"""Detecting breakish changes in timeseries value courses.

This module provides functions to detect and flag  breakish changes in the data value course, like gaps
(:py:func:`flagMissing`), jumps/drops (:py:func:`flagJumps`) or isolated values (:py:func:`flagIsolated`).
"""

from typing import Tuple

import numpy as np
import pandas as pd
import pandas.tseries.frequencies

from dios import DictOfSeries

from saqc.constants import *
from saqc.lib.tools import groupConsecutives
from saqc.funcs.changepoints import _assignChangePointCluster
from saqc.core.flags import Flags
from saqc.core.history import History
from saqc.core.register import _isflagged, register, flagging


@register(mask=[], demask=[], squeeze=["field"])
def flagMissing(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    flag: float = BAD,
    dfilter: float = FILTER_ALL,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags all values indicating missing data.

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        Container to store quality flags to data.
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The quality flags of data
    """

    datacol = data[field]
    mask = datacol.isna()

    mask = ~_isflagged(flags[field], dfilter) & mask

    flags[mask, field] = flag
    return data, flags


@flagging()
def flagIsolated(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    gap_window: str,
    group_window: str,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    The function flags arbitrary large groups of values, if they are surrounded by sufficiently
    large data gaps.

    A gap is a timespan containing either no data or data invalid only (usually `nan`) .

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The fieldname of the column, holding the data-to-be-flagged.
    flags : saqc.Flags
        A flags object
    gap_window : str
        The minimum size of the gap before and after a group of valid values, making this group considered an
        isolated group. See condition (2) and (3)
    group_window : str
        The maximum temporal extension allowed for a group that is isolated by gaps of size 'gap_window',
        to be actually flagged as isolated group. See condition (1).
    flag : float, default BAD
        flag to set.

    Returns
    -------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    flags : saqc.Flags
        The flags object, holding flags and additional information related to `data`.

    Notes
    -----
    A series of values :math:`x_k,x_{k+1},...,x_{k+n}`, with associated timestamps :math:`t_k,t_{k+1},...,t_{k+n}`,
    is considered to be isolated, if:

    1. :math:`t_{k+1} - t_n <` `group_window`
    2. None of the :math:`x_j` with :math:`0 < t_k - t_j <` `gap_window`, is valid (preceeding gap).
    3. None of the :math:`x_j` with :math:`0 < t_j - t_(k+n) <` `gap_window`, is valid (succeding gap).

    See Also
    --------
    :py:func:`flagMissing`
    """
    gap_window = pd.tseries.frequencies.to_offset(gap_window)
    group_window = pd.tseries.frequencies.to_offset(group_window)

    mask = data[field].isna()

    bools = pd.Series(data=0, index=mask.index, dtype=bool)
    for srs in groupConsecutives(mask):
        if np.all(~srs):
            # we found a chunk of non-nan values
            start = srs.index[0]
            stop = srs.index[-1]
            if stop - start <= group_window:
                # the chunk is large enough
                left = mask[start - gap_window : start].iloc[:-1]
                if left.all():
                    # the section before our chunk is nan-only
                    right = mask[stop : stop + gap_window].iloc[1:]
                    if right.all():
                        # the section after our chunk is nan-only
                        # -> we found a chunk of isolated non-values
                        bools[start:stop] = True

    flags[bools, field] = flag
    return data, flags


@flagging()
def flagJumps(
    data: DictOfSeries,
    field: str,
    flags: Flags,
    thresh: float,
    window: str,
    min_periods: int = 1,
    flag: float = BAD,
    **kwargs,
) -> Tuple[DictOfSeries, Flags]:
    """
    Flag where the mean of the values significantly changes (the data "jumps").

    Parameters
    ----------
    data : dios.DictOfSeries
        A dictionary of pandas.Series, holding all the data.
    field : str
        The reference variable, the deviation from wich determines the flagging.
    flags : saqc.Flags
        A flags object, holding flags and additional informations related to `data`.
    thresh : float
        The threshold, the mean of the values have to change by, to trigger flagging.
    window : str
        The temporal extension, of the rolling windows, the mean values that are to be
        compared, are obtained from.
    min_periods : int, default 1
        Minimum number of periods that have to be present in a window of size `window`,
        so that the mean value obtained from that window is regarded valid.
    flag : float, default BAD
        flag to set.
    """
    return _assignChangePointCluster(
        data,
        field,
        flags,
        stat_func=lambda x, y: np.abs(np.mean(x) - np.mean(y)),
        thresh_func=lambda x, y: thresh,
        window=window,
        min_periods=min_periods,
        set_flags=True,
        model_by_resids=False,
        assign_cluster=False,
        flag=flag,
        **kwargs,
    )



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import dios
import pandas as pd
import numpy as np

from saqc.funcs.noise import flagByStatLowPass
from saqc.constants import *
from saqc.core import initFlagsLike
from saqc.funcs.drift import (
    flagDriftFromNorm,
    flagDriftFromReference,
    flagDriftFromScaledNorm,
)
from saqc.funcs.outliers import flagCrossStatistic, flagRange
from saqc.funcs.flagtools import flagManual, forceFlags, clearFlags
from saqc.funcs.tools import dropField, copyField, maskTime
from saqc.funcs.resampling import concatFlags
from saqc.funcs.breaks import flagIsolated

from tests.fixtures import *
from tests.common import initData


@pytest.fixture
def data():
    return initData(cols=1, start_date="2016-01-01", end_date="2018-12-31", freq="1D")


@pytest.fixture
def field(data):
    return data.columns[0]


def test_statPass():
    data = pd.Series(0, index=pd.date_range("2000", "2001", freq="1D"), name="data")
    noise = [-1, 1] * 10
    data[100:120] = noise
    data[200:210] = noise[:10]
    data = dios.DictOfSeries(data)
    flags = initFlagsLike(data)
    data, flags = flagByStatLowPass(
        data, "data", flags, np.std, "20D", 0.999, "5D", 0.999, 0, flag=BAD
    )
    assert (flags["data"].iloc[:100] == UNFLAGGED).all()
    assert (flags["data"].iloc[100:120] == BAD).all()
    assert (flags["data"].iloc[121:] == UNFLAGGED).all()


def test_flagRange(data, field):
    min, max = 10, 90
    flags = initFlagsLike(data)
    data, flags = flagRange(data, field, flags, min=min, max=max, flag=BAD)
    flagged = flags[field] > UNFLAGGED
    expected = (data[field] < min) | (data[field] > max)
    assert all(flagged == expected)


def test_flagSesonalRange(data, field):
    data.iloc[::2] = 0
    data.iloc[1::2] = 50
    nyears = len(data[field].index.year.unique())

    tests = [
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 7,
                "startday": 1,
                "endmonth": 8,
                "endday": 31,
            },
            31 * 2 * nyears // 2,
        ),
        (
            {
                "min": 1,
                "max": 100,
                "startmonth": 12,
                "startday": 16,
                "endmonth": 1,
                "endday": 15,
            },
            31 * nyears // 2 + 1,
        ),
    ]

    for test, expected in tests:
        flags = initFlagsLike(data)
        newfield = f"{field}_masked"
        start = f"{test['startmonth']:02}-{test['startday']:02}T00:00:00"
        end = f"{test['endmonth']:02}-{test['endday']:02}T00:00:00"

        data, flags = copyField(data, field, flags, field + "_masked")
        data, flags = maskTime(
            data,
            newfield,
            flags,
            mode="periodic",
            start=start,
            end=end,
            closed=True,
            flag=BAD,
        )
        data, flags = flagRange(
            data, newfield, flags, min=test["min"], max=test["max"], flag=BAD
        )
        data, flags = concatFlags(
            data, newfield, flags, method="match", target=field, flag=BAD
        )
        data, flags = dropField(data, newfield, flags)
        flagged = flags[field] > UNFLAGGED
        assert flagged.sum() == expected


def test_clearFlags(data, field):
    flags = initFlagsLike(data)
    flags[:, field] = BAD
    assert all(flags[field] == BAD)

    _, flags = clearFlags(data, field, flags)
    assert all(flags[field] == UNFLAGGED)


def test_forceFlags(data, field):
    flags = initFlagsLike(data)
    flags[:, field] = BAD
    assert all(flags[field] == BAD)

    _, flags = forceFlags(data, field, flags, flag=DOUBTFUL)
    assert all(flags[field] == DOUBTFUL)


def test_flagIsolated(data, field):
    flags = initFlagsLike(data)

    data.iloc[1:3, 0] = np.nan
    data.iloc[4:5, 0] = np.nan
    flags[data[field].index[5:6], field] = BAD
    data.iloc[11:13, 0] = np.nan
    data.iloc[15:17, 0] = np.nan

    #              data  flags
    # 2016-01-01   0.0   -inf
    # 2016-01-02   NaN   -inf
    # 2016-01-03   NaN   -inf
    # 2016-01-04   3.0   -inf
    # 2016-01-05   NaN   -inf
    # 2016-01-06   5.0  255.0
    # 2016-01-07   6.0   -inf
    # 2016-01-08   7.0   -inf
    #         ..    ..     ..

    _, flags_result = flagIsolated(
        data, field, flags, group_window="1D", gap_window="2.1D", flag=BAD
    )

    assert flags_result[field].iloc[[3, 5]].all()

    data, flags_result = flagIsolated(
        data,
        field,
        flags_result,
        group_window="2D",
        gap_window="2.1D",
        continuation_range="1.1D",
        flag=BAD,
    )
    assert flags_result[field].iloc[[3, 5, 13, 14]].all()


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_2")])
def test_flagCrossScoring(dat):
    data1, characteristics = dat(initial_level=0, final_level=0, out_val=0)
    data2, characteristics = dat(initial_level=0, final_level=0, out_val=10)
    fields = ["data1", "data2"]
    s1, s2 = data1.squeeze(), data2.squeeze()
    s1 = pd.Series(data=s1.values, index=s1.index)
    s2 = pd.Series(data=s2.values, index=s1.index)
    data = dios.DictOfSeries([s1, s2], columns=["data1", "data2"])
    flags = initFlagsLike(data)
    _, flags_result = flagCrossStatistic(
        data, fields, flags, thresh=3, method=np.mean, flag=BAD
    )
    for field in fields:
        isflagged = flags_result[field] > UNFLAGGED
        assert isflagged[characteristics["raise"]].all()


def test_flagManual(data, field):
    flags = initFlagsLike(data)
    dat = data[field]

    mdata = pd.Series("lala", index=dat.index)
    index_exp = mdata.iloc[[10, 33, 200, 500]].index
    mdata.iloc[[101, 133, 220, 506]] = "b"
    mdata.loc[index_exp] = "a"
    shrinked = mdata.loc[index_exp.union(mdata.iloc[[1, 2, 3, 4, 600, 601]].index)]

    kwargs_list = [
        dict(mdata=mdata, mflag="a", method="plain", mformat="mflag", flag=BAD),
        dict(mdata=mdata, mflag="a", method="ontime", mformat="mflag", flag=BAD),
        dict(mdata=shrinked, mflag="a", method="ontime", mformat="mflag", flag=BAD),
    ]

    for kw in kwargs_list:
        _, fl = flagManual(data.copy(), field, flags.copy(), **kw)
        isflagged = fl[field] > UNFLAGGED
        assert isflagged[isflagged].index.equals(index_exp)

    # flag not exist in mdata
    _, fl = flagManual(
        data.copy(),
        field,
        flags.copy(),
        mdata=mdata,
        mflag="i do not exist",
        method="ontime",
        mformat="mflag",
        flag=BAD,
    )
    isflagged = fl[field] > UNFLAGGED
    assert isflagged[isflagged].index.equals(pd.DatetimeIndex([]))

    # check closure methods
    index = pd.date_range(start="2016-01-01", end="2018-12-31", periods=11)
    mdata = pd.Series(0, index=index)
    mdata.loc[index[[1, 5, 6, 7, 9, 10]]] = 1
    # >>> mdata
    # 2016-01-01 00:00:00    0
    # 2016-04-19 12:00:00    1
    # 2016-08-07 00:00:00    0
    # 2016-11-24 12:00:00    0
    # 2017-03-14 00:00:00    0
    # 2017-07-01 12:00:00    1
    # 2017-10-19 00:00:00    1
    # 2018-02-05 12:00:00    1
    # 2018-05-26 00:00:00    0
    # 2018-09-12 12:00:00    1
    # 2018-12-31 00:00:00    1
    # dtype: int64
    m_index = mdata.index
    flag_intervals = [
        (m_index[1], m_index[2]),
        (m_index[5], m_index[8]),
        (m_index[9], dat.index.shift(freq="1h")[-1]),
    ]
    bound_drops = {"right-open": [1], "left-open": [0], "closed": []}
    for method in ["right-open", "left-open", "closed"]:
        _, fl = flagManual(
            data.copy(),
            field,
            flags.copy(),
            mdata=mdata,
            mflag=1,
            method=method,
            mformat="mflag",
            flag=BAD,
        )
        isflagged = fl[field] > UNFLAGGED
        for flag_i in flag_intervals:
            f_i = isflagged[slice(flag_i[0], flag_i[-1])].index
            check_i = f_i.drop(
                [flag_i[k] for k in bound_drops[method]], errors="ignore"
            )
            assert isflagged[check_i].all()
            unflagged = isflagged[f_i.difference(check_i)]
            if not unflagged.empty:
                assert ~unflagged.all()


@pytest.mark.parametrize("dat", [pytest.lazy_fixture("course_1")])
def test_flagDriftFromNormal(dat):
    data = dat(periods=200, peak_level=5, name="d1")[0]
    data["d2"] = dat(periods=200, peak_level=10, name="d2")[0]["d2"]
    data["d3"] = dat(periods=200, peak_level=100, name="d3")[0]["d3"]
    data["d4"] = 3 + 4 * data["d1"]
    data["d5"] = 3 + 4 * data["d1"]

    flags = initFlagsLike(data)
    _, flags_norm = flagDriftFromNorm(
        data=data.copy(),
        field=["d1", "d2", "d3"],
        flags=flags.copy(),
        freq="200min",
        spread=5,
        flag=BAD,
    )

    _, flags_ref = flagDriftFromReference(
        data=data.copy(),
        field=["d1", "d2", "d3"],
        flags=flags.copy(),
        reference="d1",
        freq="3D",
        thresh=20,
        flag=BAD,
    )

    _, flags_scale = flagDriftFromScaledNorm(
        data.copy(),
        "dummy",
        flags.copy(),
        ["d1", "d3"],
        ["d4", "d5"],
        freq="3D",
        thresh=20,
        spread=5,
        flag=BAD,
    )
    assert all(flags_norm["d3"] > UNFLAGGED)
    assert all(flags_ref["d3"] > UNFLAGGED)
    assert all(flags_scale["d3"] > UNFLAGGED)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import annotations

from typing import Sequence, Callable, Optional, Union
from typing_extensions import Literal

import numpy as np
from scipy.spatial.distance import pdist

from saqc.constants import BAD
import saqc
from saqc.funcs import LinkageString
from saqc.lib.types import CurveFitter


class Drift:
    def flagDriftFromNorm(
        self,
        field: Sequence[str],
        freq: str,
        spread: float,
        frac: float = 0.5,
        metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
            np.array([x, y]), metric="cityblock"
        )
        / len(x),
        method: LinkageString = "single",
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagDriftFromNorm", locals())

    def flagDriftFromReference(
        self,
        field: Sequence[str],
        reference: str,
        freq: str,
        thresh: float,
        metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
            np.array([x, y]), metric="cityblock"
        )
        / len(x),
        target=None,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagDriftFromReference", locals())

    def flagDriftFromScaledNorm(
        self,
        field: str,
        set_1: Sequence[str],
        set_2: Sequence[str],
        freq: str,
        spread: float,
        frac: float = 0.5,
        metric: Callable[[np.ndarray, np.ndarray], float] = lambda x, y: pdist(
            np.array([x, y]), metric="cityblock"
        )
        / len(x),
        method: LinkageString = "single",
        target: str = None,
        flag: float = BAD,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("flagDriftFromScaledNorm", locals())

    def correctDrift(
        self,
        field: str,
        maintenance_field: str,
        model: Callable[..., float] | Literal["linear", "exponential"],
        cal_range: int = 5,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("correctDrift", locals())

    def correctRegimeAnomaly(
        self,
        field: str,
        cluster_field: str,
        model: CurveFitter,
        tolerance: Optional[str] = None,
        epoch: bool = False,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("correctRegimeAnomaly", locals())

    def correctOffset(
        self,
        field: str,
        max_jump: float,
        spread: float,
        window: str,
        min_periods: int,
        tolerance: Optional[str] = None,
        **kwargs
    ) -> saqc.SaQC:
        return self._defer("correctOffset", locals())



from .test_setup import *
import pytest

#
# s1 = pd.Series(range(10), index=range(10))
# s2 = pd.Series(range(5, 10), index=range(5, 10))
# s3 = pd.Series(range(1, 30, 2), index=range(1, 30, 2))
# s4 = pd.Series(np.linspace(7, 13, 9), index=range(3, 12))
# s1.name, s2.name, s3.name, s4.name = 'a', 'b', 'c', 'd'
# d1 = DictOfSeries(data=dict(a=s1.copy(), b=s2.copy(), c=s3.copy(), d=s4.copy()))
#
# blist = [True, False, False, True]
# b = pd.Series([True, False] * 5, index=[1, 2, 3, 4, 5] + [6, 8, 10, 12, 14])
# B = d1 > 5
#
#
#
#
# BLIST = [True, False, False, True]
#
# LISTIDXER = [['a'], ['a', 'c'], pd.Series(['a', 'c'])]
# BOOLIDXER = [pd.Series(BLIST), d1.copy() > 10]
# SLICEIDXER = [slice(None), slice(-3, -1), slice(-1, 3), slice(None, None, 3)]
# MULTIIDXER = []  # [d1 > 9, d1 != d1, d1 == d1]
# EMPTYIDEXER = [[], pd.Series(), slice(3, 3), slice(3, -1), DictOfSeries()]
#
# INDEXERS = LISTIDXER + BOOLIDXER + SLICEIDXER + MULTIIDXER + EMPTYIDEXER
#
#



from setuptools import setup, find_packages
import subprocess


with open("Readme.md", "r") as fh:
    long_description = fh.read()

cmd = "git describe --tags --always --dirty"
version = (
    subprocess.run(cmd, shell=True, check=False, stdout=subprocess.PIPE)
    .stdout.decode()
    .strip()
)
print(f"git version: {version}")
# if '-dirty' in version:
#     print("Do not make a version from a dirty repro. Exiting now")
#     exit(1)
txt = "enter version\n>"
version = input(txt)

setup(
    name="dios",
    version=version,
    author="Bert Palm",
    author_email="bert.palm@ufz.de",
    description="Dictionary of Series - a kind of pandas extension",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://git.ufz.de/rdm/dios",
    packages=["dios"],
    install_requires=[
        "pandas",
    ],
    license="GPLv3",
)
